/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "VariantWriter.h"

#include <memory>
#include <numeric>

#include <RobotAPI/libraries/aron/core/type/variant/All.h>
#include <RobotAPI/libraries/aron/core/type/visitor/variant/VariantVisitor.h>

namespace armarx::aron::type::writer
{

    type::Descriptor
    VariantWriter::getDescriptor(ReturnTypeConst& input)
    {
        return ConstVariantVisitor::GetDescriptor(input);
    }

    aron::type::VariantPtr
    VariantWriter::writeObject(const std::string& name,
                               const std::vector<std::string>& templates,
                               const std::vector<std::string>& templateInstantiations,
                               const std::map<std::string, aron::type::VariantPtr>& memberTypes,
                               const std::optional<aron::type::VariantPtr>& extends,
                               const type::Maybe maybe,
                               const Path& p)
    {
        auto o = std::make_shared<type::Object>(p);
        o->setObjectName(name);
        o->setTemplates(templates);
        o->setTemplateInstantiations(templateInstantiations);
        o->setMemberTypes(memberTypes);
        o->setMaybe(maybe);
        if (extends.has_value())
        {
            auto ex = type::Object::DynamicCast(*extends);
            o->setExtends(ex);
        }
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeList(const aron::type::VariantPtr& acceptedType,
                             const type::Maybe maybe,
                             const Path& p)
    {
        auto o = std::make_shared<type::List>(p);
        o->setAcceptedType(acceptedType);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeDict(const aron::type::VariantPtr& acceptedType,
                             const type::Maybe maybe,
                             const Path& p)
    {
        auto o = std::make_shared<type::Dict>(p);
        o->setAcceptedType(acceptedType);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writePair(const aron::type::VariantPtr& acceptedType1,
                             const aron::type::VariantPtr& acceptedType2,
                             const type::Maybe maybe,
                             const Path& p)
    {
        auto o = std::make_shared<type::Pair>(p);
        o->setFirstAcceptedType(acceptedType1);
        o->setSecondAcceptedType(acceptedType2);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeTuple(const std::vector<aron::type::VariantPtr>& acceptedTypes,
                              const type::Maybe maybe,
                              const Path& p)
    {
        auto o = std::make_shared<type::Tuple>(p);
        o->setAcceptedTypes(acceptedTypes);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeNDArray(const int ndim,
                                const type::ndarray::ElementType type,
                                const std::string& defaultValue,
                                const type::Maybe maybe,
                                const Path& p)
    {
        auto o = std::make_shared<type::NDArray>(p);
        o->setMaybe(maybe);
        o->setElementType(type);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeMatrix(const int rows,
                               const int cols,
                               const type::matrix::ElementType type,
                               const std::string& defaultValue,
                               const type::Maybe maybe,
                               const Path& p)
    {
        auto o = std::make_shared<type::Matrix>(p);
        o->setMaybe(maybe);
        o->setRows(rows);
        o->setCols(cols);
        o->setElementType(type);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeQuaternion(const type::quaternion::ElementType type,
                                   const std::string& defaultValue,
                                   const type::Maybe maybe,
                                   const Path& p)
    {
        auto o = std::make_shared<type::Quaternion>(p);
        o->setMaybe(maybe);
        o->setElementType(type);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeImage(const type::image::PixelType type,
                              const std::string& defaultValue,
                              const type::Maybe maybe,
                              const Path& p)
    {
        auto o = std::make_shared<type::Image>(p);
        o->setMaybe(maybe);
        o->setPixelType(type);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writePointCloud(const type::pointcloud::VoxelType type,
                                   const std::string& defaultValue,
                                   const type::Maybe maybe,
                                   const Path& p)
    {
        auto o = std::make_shared<type::PointCloud>(p);
        o->setMaybe(maybe);
        o->setVoxelType(type);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeIntEnum(const std::string& name,
                                const std::map<std::string, int>& acceptedValues,
                                const std::string& defaultValue,
                                const type::Maybe maybe,
                                const Path& p)
    {
        auto o = std::make_shared<type::IntEnum>(p);
        o->setEnumName(name);
        o->setAcceptedValueMap(acceptedValues);
        o->setDefaultValueName(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeInt(const std::optional<int>& defaultValue,
                            const type::Maybe maybe,
                            const Path& p)
    {
        auto o = std::make_shared<type::Int>(p);
        o->setDefaultValue(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeLong(const std::optional<long>& defaultValue,
                             const type::Maybe maybe,
                             const Path& p)
    {
        auto o = std::make_shared<type::Long>(p);
        o->setDefaultValue(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeFloat(const std::optional<float>& defaultValue,
                              const type::Maybe maybe,
                              const Path& p)
    {
        auto o = std::make_shared<type::Float>(p);
        o->setDefaultValue(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeDouble(const std::optional<double>& defaultValue,
                               const type::Maybe maybe,
                               const Path& p)
    {
        auto o = std::make_shared<type::Double>(p);
        o->setDefaultValue(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeString(const std::optional<std::string>& defaultValue,
                               const type::Maybe maybe,
                               const Path& p)
    {
        auto o = std::make_shared<type::String>(p);
        o->setDefaultValue(defaultValue);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeBool(const std::optional<bool>& defaultValue,
                             const type::Maybe maybe,
                             const Path& p)
    {
        auto o = std::make_shared<type::Bool>(p);
        o->setMaybe(maybe);
        o->setDefaultValue(defaultValue);
        return o;
    }

    aron::type::VariantPtr
    VariantWriter::writeAnyObject(const type::Maybe maybe, const Path& p)
    {
        auto o = std::make_shared<type::AnyObject>(p);
        o->setMaybe(maybe);
        return o;
    }
} // namespace armarx::aron::type::writer
